/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <memory.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <mplib1/mplib1_config.h>

#include "../include/bpo_internal.h"
#include <mplib1/bpo.h>

#include <mplib1/safe_string.h>
#include <mplib1/dl_list.h>
#include <mplib1/cfg_file.h>
#include <mplib1/fprintfile.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    structures
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    static variables
   ------------------------------------------------------------------ */

static int done_init=0;
static int resource_internal=0;
static int res_debug=0;
static const char bpo_resource_debug[]="BPO_RESOURCE_DEBUG";

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

void *
Find_SODB_Resource( const void *hint, const char *res_name )
{
    struct sodb *sodbp;
    struct sodb_resource *sodbrp;
    void *rv=NULL;

    res_debug = get_config_flag(bpo_resource_debug);
    if (res_debug)
	fprintfile( stderr,"Find_SODB_Resource: hint: %p  Name: %s\n",
		    hint, res_name );
    if (done_init==0)
    {
	Do_SODB_Init();
	done_init=1;
    }
    sodbp = Get_SODB_Base( hint );
    if (sodbp != NULL)
    {
	/* Special case, check for base name */
	if (strcmp( res_name, sodbp->sodb_res_name )==0)
	{
	    rv = sodbp;
	}else
	{
	    sodbrp = bpo_Find_Item_By_Name( &sodbp->sodb_resource_list,
					    res_name );
	    if (sodbrp)
		rv = GPTR(sodbp,sodbrp->res_off);
	}
    }
    if (res_debug)
	fprintfile( stderr,"Find_SODB_Resource: rv=%p\n", rv );
    return(rv);
}

static off_t
add_res_func( bpo_List_t *list, struct sodb_res_find *resfp )
{
/*    struct sodb *sodbp;*/
    struct sodb_resource *sodbrp;
    const char *cp;
    size_t tsize;
    int no_name;
    off_t rv=0,*ofp;

    sodbrp = bpo_Find_Item_By_Name( list, resfp->name );
    if (sodbrp)
    {
	resfp->ierrno = EEXIST;
    }else
    {
	if (   (resfp->flags & SODB_RES_NO_ALLOC_NM)
	    && SODB_in_Segment( list, resfp->name )
	   )
	{
	    tsize = sizeof(struct sodb_resource) + sizeof(off_t);
	    no_name=1;
	}else
	{
	    tsize = sizeof(struct sodb_resource) + sizeof(off_t)
		    + Sstrlen(resfp->name) + 1;
	    no_name=0;
	}
	resource_internal = 1;
	ofp = shalloc( resfp->sodbp, (size_t)tsize );
	resource_internal = 0;
	if (ofp)
	{
	    sodbrp = (struct sodb_resource *)(ofp+1);
	    /* Got it, so init it */
	    if (no_name)
	    {
		cp = resfp->name;
	    }else
	    {
		Sstrcpy( (char *)sodbrp + sizeof(struct sodb_resource),
			resfp->name );
		cp = ((char *)sodbrp) + sizeof(struct sodb_resource);
	    }
	    sodbrp->res_off = GOFF( resfp->sodbp, resfp->obj );
	    bpo_Init_Node( &sodbrp->res_node, cp, sodbrp );
	    bpo_Add_Tail( list, &sodbrp->res_node );
	    resfp->rv = GOFF(resfp->sodbp, sodbrp);
	}
    }
    return( rv );
}

off_t
Add_SODB_Resource( void *object, const char *res_name )
{
    struct sodb_res_find res_find;
    struct sodb *sodbp;
    off_t rv=0;

    res_debug = get_config_flag(bpo_resource_debug);
    if (res_debug)
	fprintfile( stderr,"Add_SODB_Resource: Object: %p  Name: %s\n",
		    object, res_name );
    if (done_init==0)
	Do_SODB_Init();
    sodbp = Get_SODB_Base( object );
    if (sodbp && (sodbp->sodb_flags & SODB_USE_RESOURCES))
    {
	res_find.sodbp = sodbp;
	res_find.name = res_name;
	res_find.ptr = NULL;
	res_find.obj = object;
	res_find.flags = 0;
	res_find.ierrno = 0;
	res_find.rv = (off_t)0;
	bpo_Work_List( &sodbp->sodb_resource_list,
			(bpo_Work_List_t)add_res_func,
			&res_find );

	errno = res_find.ierrno;
	rv = res_find.rv;
    }
    if (res_debug)
	fprintfile( stderr,"Add_SODB_Resource: rv=%ld\n", (int)rv );
    return( rv );
}


void
Init_SODB_Resources( struct sodb *sodbp )
{
    bpo_Init_pid_lock( &sodbp->res_p_lock, NULL );
    bpo_pid_Lock( &sodbp->res_p_lock );

    bpo_Init_List_Raw( &sodbp->sodb_resource_list, sodbp, LN_IGNORECASE );
    Sstrcpy( sodbp->sodb_res_name, SODB_BASE_NAME );
    sodbp->sodb_resource.res_off = 0;
    bpo_Init_Node_Raw( &sodbp->sodb_resource.res_node, sodbp,
			sodbp->sodb_res_name, &sodbp->sodb_resource );
    bpo_Add_Head_Raw( &sodbp->sodb_resource_list, sodbp,
			&sodbp->sodb_resource.res_node );

    bpo_pid_Unlock( &sodbp->res_p_lock );

    return;
}


void *
Fetch_SODB_Resource( const void *hint, const char *name, size_t s_size,
		    Fetch_SODB_Resource_t s_init )
{
    void *rv,*nrv;
    off_t noff, *ofp;

    rv = Find_SODB_Resource( hint, name );
    if (rv==NULL)
    {
        ofp = shalloc( hint, s_size + sizeof(off_t) );
        if (ofp)
        {
            *ofp = 0;
            nrv = (void *)(ofp+1);
            if (s_init)
		(* s_init)( nrv );
            noff = Add_SODB_Resource( nrv, name );
            if (noff== (off_t)0)
            {
                shfree(ofp);
                nrv=NULL;
            }
            rv = Find_SODB_Resource( hint, name );
        }
    }
    return(rv);
}


/* -- End of File -- */
