/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#ifdef INCLUDE_GETOPT
#include <getopt.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <memory.h>
#include <unistd.h>

#include <mplib1/mplib1_config.h>

#include "../include/bpo_internal.h"
#include <mplib1/bpo.h>

#include <mplib1/cfg_file.h>
#include <mplib1/dl_list.h>
#include <mplib1/fprintfile.h>
#include <mplib1/pid_check.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

#define GENOFF(a,b)     (off_t)((char *)a-(char *)b)

/* ------------------------------------------------------------------
    structures
   ------------------------------------------------------------------ */

struct pwalk_res
{
int		which_procs;
Validate_Resource_t	isp_disp;
struct	sodb	*sodbp;
void		*f_param1;
};

struct pwalk_proc
{
int	which_procs;
int (*isp_disp)( struct shm_process_private *shmpp, void *param1 );
struct	shm_proc_list	*w_splp;
struct cache_proc_list	*w_cpl;
void			*f_param1;
};

struct pwalk_q
{
int		which_queues;
int (*isp_disp)( struct bpo_private_q *pqp, void *param1 );
struct	Q_Head	*qh;
void		*f_param1;
};

/* ------------------------------------------------------------------
    static variables
   ------------------------------------------------------------------ */

static int (*v_func)(FILE *,const char *,...)=NULL;
static FILE *v_fh=NULL;
static int list_has_problems=0;

static dl_List_t del_list;
static int n_del;

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

int
node_list_header( int (*p_func)(FILE *,const char *,...), FILE *fh )
{
    int rv=1;
    if (p_func)
    rv = (* p_func)( fh, "Address  lno_me    Forward  Backward List     Item     Name\n" );
    return(rv);
}

int
node_list_details( void *base, bpo_Node_t *bnp,
		    int (*p_func)(FILE *,const char *,...),
		    FILE *fh )

{
    int rv=1;
    if (p_func)
    rv = (*p_func)( fh, "%08lx(%08lx) %08lx %08lx %08lx %08lx <%s>\n",
		    GENOFF(bnp,base), bnp->lno_me,
		    bnp->lno_Succ, bnp->lno_Pred,
		    bnp->lno_List, bnp->lno_Item,
		    (bnp->lno_Name)? (char *)base + bnp->lno_Name : "" );
    return(rv);
}

static int
node_details( void *vp, bpo_Node_t *bnp, int dbg_lev, int error_level )
{
    if (error_level==0 && dbg_lev==0)
	node_list_header( v_func, v_fh );
    if (dbg_lev==0)
	node_list_details( vp, bnp, v_func, v_fh );
    return(error_level+1);
}

static int
Walk_Forwards( void *vp, bpo_List_t *blp, bpo_Node_t *bnp, int depth )
{
    off_t as_should;
    bpo_Node_t *nnp;
    int rv=0;

    /* check the various aspects of this node */
    as_should = (off_t)((char *)bnp - (char *)vp);

    /* does the node have a valid 'me' value */
    if (bnp->lno_me != as_should)
    {
	rv = node_details( vp, bnp, 0, rv );
	if (v_func)
	    (* v_func)( v_fh, "    ^  vs   ^    Bad lno_me\n" );
	list_has_problems++;
    }

    /* Is the Node actually in the list it's meant to be in? */
    if (bnp->lno_List != blp->lno_me)
    {
	rv = node_details( vp, bnp, 0, rv );
	if (v_func)
	    (* v_func)(v_fh, " node is not in list %08lx(%08lx)\n",
		    GENOFF(blp,vp), blp->lno_me );
	list_has_problems++;
    }
    if (bnp->lno_Succ)
    {
	/* the next node pointed to should point back to us */
	nnp = GRPTR(bnp,bnp->lno_Succ);
	if (nnp->lno_Pred != bnp->lno_me)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(* v_func)( v_fh, " successor does not point back to us\n" );
	    node_list_details( vp, nnp, v_func, v_fh );
	    list_has_problems++;
	}
	/* Time to recurse for the next node */
	if (rv == 0)
	    rv += Walk_Forwards( vp, blp, nnp, depth+1 );
    }else
    {
	/* well in theory, this is the tail node in the list structure, so check it */
	if (bnp != &blp->lno_Tail)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(* v_func)( v_fh, " node with no successor is not tail node %08lx(%08lx)\n",
			    GENOFF(&blp->lno_Tail,vp), blp->lno_Tail.lno_me );
	    list_has_problems++;
	}
    }
    if (depth)
    {
	/* check back link */
	nnp = GRPTR(bnp,bnp->lno_Pred);
	if (nnp->lno_Succ != bnp->lno_me)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(*v_func)( v_fh, " predecessor does not point to me %08lx -> %08lx\n",
			GENOFF(nnp,vp), nnp->lno_Succ );
	    list_has_problems++;
	}
    }
    return(rv);
}

static int
Walk_Backwards( void *vp, bpo_List_t *blp, bpo_Node_t *bnp, int depth )
{
    off_t as_should;
    bpo_Node_t *nnp;
    int rv=0;

    /* check the various aspects of this node */
    as_should = (off_t)((char *)bnp - (char *)vp);

    if (0)
	node_list_details( vp, bnp, v_func, v_fh );
    /* does the node have a valid 'me' value */
    if (bnp->lno_me != as_should)
    {
	rv = node_details( vp, bnp, 0, rv );
	if (v_func)
	    (* v_func)( v_fh, "    ^  vs   ^    Bad lno_me\n" );
	list_has_problems++;
    }

    /* Is the Node actually in the list it's meant to be in? */
    if (bnp->lno_List != blp->lno_me)
    {
	rv = node_details( vp, bnp, 0, rv );
	if (v_func)
	    (*v_func)( v_fh, " node is not in list %08lx(%08lx)\n",
		    GENOFF(blp,vp), blp->lno_me );
	list_has_problems++;
    }
    if (bnp->lno_Pred)
    {
	/* the next node pointed to should point back to us */
	nnp = GRPTR(bnp,bnp->lno_Pred);
	if (nnp->lno_Succ != bnp->lno_me)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(*v_func)( v_fh, " predecessor does not point to me %08lx -> %08lx\n",
			GENOFF(nnp,vp), nnp->lno_Pred );
	    list_has_problems++;
	}
	/* Time to recurse for the next node */
	if (rv == 0)
	    rv += Walk_Backwards( vp, blp, nnp, depth+1 );
    }else
    {
	/* well in theory, this is the tail node in the list structure, so check it */
	if (bnp != &blp->lno_Head)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(*v_func)( v_fh, " node with no predecessor is not head node %08lx(%08lx)\n",
			GENOFF(&blp->lno_Head,vp), blp->lno_Head.lno_me );
	    list_has_problems++;
	}
    }
    if (depth)
    {
	/* check forw link */
	nnp = GRPTR(bnp,bnp->lno_Succ);
	if (nnp->lno_Pred != bnp->lno_me)
	{
	    rv = node_details( vp, bnp, 0, rv );
	    if (v_func)
		(*v_func)( v_fh, " successor does not point back to us\n" );
	    node_list_details( vp, nnp, v_func, v_fh );
	    list_has_problems++;
	}
    }
    return(rv);
}


int
bpo_List_Chk( bpo_List_t *rlp,
		int (*p_func)(FILE *,const char *,...),
		FILE *fh )
{
    int rv=0;
    void *vp;

    v_func = p_func;
    v_fh = fh;
    list_has_problems=0;

    vp = Get_SODB_Base( rlp );

    if(vp)
    {
	if (   (rlp->lno_me != ((char *)rlp - (char *)vp) )
	    || (rlp->lno_Head.lno_me != ((char *)(&rlp->lno_Head) - (char *)vp) )
	    || (rlp->lno_Tail.lno_me != ((char *)(&rlp->lno_Tail) - (char *)vp) )
	    || (rlp->lno_Head.lno_List != rlp->lno_me)
	    || (rlp->lno_Tail.lno_List != rlp->lno_me)
	   )
	{
	    if (v_func)
		(*v_func)( v_fh, "list header does not contain correct offset\n" );
	    list_has_problems++;
	}
	else
	{
	    bpo_pid_Lock( &rlp->lno_Pid_Lock );
/*	    list_top_fix( vp, rlp );*/
	    rv = Walk_Forwards( vp, rlp, &rlp->lno_Head, 0 );
	    if (rv)
		rv = Walk_Backwards( vp, rlp, &rlp->lno_Tail, 0 );
	    bpo_pid_Unlock( &rlp->lno_Pid_Lock );
	}
    }else
    {
	if (v_func)
	    (* v_func)( v_fh, "list is not is shared block\n" );
	list_has_problems++;
    }
    return(list_has_problems);
}

int
bpo_Validate_List( bpo_List_t *rlp )
{
    return (bpo_List_Chk( rlp, fprintf, stderr ));
}



static void
shm_check_proc( struct shm_process_private *sppp, struct pwalk_proc *pw )
{
    int rv = SHM_LEAVE_PROC;

    switch(pw->which_procs)
    {
    case SHM_CHECK_ALL_PROCESSES :
	rv = (* pw->isp_disp)( sppp, pw->f_param1 );
        break;
    case SHM_CHECK_DEAD_PROCESSES :
	if (is_pid_dead( sppp->pid ) != 0)
	{
	    rv = (* pw->isp_disp)( sppp, pw->f_param1 );
	}
	break;
    default :
	/* huh? */
	break;
    }

    if (rv == SHM_DELETE_PROC)
    {
	Remove_Shm_Process( pw->w_splp, sppp->pid_str );
    }
    return;
}

int
Validate_Processes( const void * hint,
		    int p_which_procs,
		    Validate_Processes_t disp,
		    void *param1 )
{
    /* OK, so walk the process list and call the routine for each process that is dead */
    struct shm_proc_list *splp;
    struct shm_process_private *sppp;
    struct cache_proc_list *cpl;
    struct pwalk_proc myprocwalk;
    int rv=0;

/*    init_pid_stuff();*/

    /* First we need to ensure that the process list resource exists */
    splp = Find_Proc_List( hint );
    if ( splp == NULL )
	return(0);
    cpl = Get_Cache_Procs( splp );
    if (cpl==NULL)
	return(0);

    myprocwalk.which_procs = p_which_procs;
    myprocwalk.w_splp = splp;
    myprocwalk.w_cpl = cpl;
    myprocwalk.isp_disp = disp;
    myprocwalk.f_param1 = param1;

    if (p_which_procs == SHM_CHECK_THIS_PROCESS )
    {
	sppp = bpo_Find_Item_By_Name( &splp->proc_List, param1 );
	if (sppp)
	{
	    if (disp)
		rv = (* disp)( sppp, param1 );
	    else rv = 1;
	}
	else rv = -1;
    }else
    {
	dl_Walk_List( &cpl->proc_list,
		    (dl_Walk_List_t)shm_check_proc,
		    &myprocwalk );
    }
    return(rv);
}


static int
do_del_queues( bpo_List_t *blp, dl_List_t *lp )
{
    struct bpo_private_q *pqp;

    while( (pqp = dl_Remove_Head_Item( lp )) )
    {
	/* Now, since we have the real queueu list locked we can check
	   whether this queue is still in the list. If it is, we
	   remove, and free the queue. If not, we assume that the
	   queue has been freed by someone else and all we need to do
	   is to free up our local details
	*/
	if (bpo_Find_Node_By_Item( blp, pqp->qp )!=NULL)
	{
	    bpo_Remove_Node( &pqp->qp->Q_Node );
	    /* We now have a dissociated qp, all we have to do is loose
	       all the associated items
	    */
	    free_q_stuff( pqp->qp );
	}
	free_this_private_q( pqp );
    }
    return(0);
}


static void
q_check_proc( struct bpo_private_q *pqp, struct pwalk_q *pw )
{
    int rv = Q_LEAVE_QUEUE;

    switch(pw->which_queues)
    {
    case Q_CHECK_ALL_QUEUES :
	rv = (* pw->isp_disp)( pqp, pw->f_param1 );
        break;
    case Q_CHECK_DEAD_QUEUES :
	if (   (pqp->Q_flags & Q_SOLE_READER)
	    && (pqp->Q_flags & Q_PERSISTENT)==0
	    && is_pid_dead( pqp->Q_sole_pid ) != 0)
	{
	    rv = (* pw->isp_disp)( pqp, pw->f_param1 );
	}
	break;
    default :
	/* huh? */
	break;
    }

    if (rv == Q_DELETE_QUEUE)
    {
	dl_Remove_Node( &pqp->q_Node );
	dl_Add_Tail( &del_list, &pqp->q_Node );
	n_del++;
    }
    return;
}

int
Validate_Queues( const void * hint,
		int p_which_queues,
		Validate_Queues_t disp,
		void *param1 )
{
    /* OK, so walk the process list and call the routine for each process that is dead */
    struct Q_Head *qh;
    struct pwalk_q myqwalk;
    struct cache_queue_list *qpl;

    qh = get_Q_resource( hint );
    if (qh==NULL)
    {
	errno = ENOSYS;
	return(-1);
    }

    dl_Init_List( &del_list, 0 );
    n_del=0;

    myqwalk.which_queues = p_which_queues;
    myqwalk.qh = qh;
    myqwalk.isp_disp = disp;
    myqwalk.f_param1 = param1;

    qpl = cache_queues( qh, NULL );

    dl_Walk_List( &qpl->q_list, (dl_Walk_List_t)q_check_proc, &myqwalk );
    if (n_del)
	bpo_Work_List( &qh->Qh_List,
			(bpo_Work_List_t)do_del_queues,
			&del_list );
    return(0);
}

static void
res_check_proc( void *vp1, void *vp2 )
{
    struct sodb_resource *sodbrp;
    struct pwalk_res *pw;
    char *res_name;

    sodbrp = vp1;
    pw = vp2;
    res_name = GRPTR( &sodbrp->res_node.lno_me, sodbrp->res_node.lno_Name );

    (* pw->isp_disp)( res_name, sodbrp->res_off, pw->f_param1 );
    return;
}

int
Validate_Resources( const void * hint,
		    int p_which_procs,
		    Validate_Resource_t disp,
		    void *param1 )
{
    /* OK, so walk the resource list and call the routine for each
       resource that we find */
    struct sodb *sodbp;
    struct pwalk_res myprocwalk;
    int rv=0;

    sodbp = Get_SODB_Base( hint );
    if (sodbp == NULL)
	return(rv);

    myprocwalk.which_procs = p_which_procs;
    myprocwalk.sodbp = sodbp;
    myprocwalk.isp_disp = disp;
    myprocwalk.f_param1 = param1;

    bpo_pid_Lock( &sodbp->res_p_lock );
    bpo_Walk_List( &sodbp->sodb_resource_list, res_check_proc, &myprocwalk );
    bpo_pid_Unlock( &sodbp->res_p_lock );
    return(0);
}

/* -- End of File -- */
