/*
*******************************************************************************
*  Copyright (c) 1997 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <signal.h>

#include <mplib1/dl_list.h>
#include <mplib1/pid_check.h>
#include <mplib1/cfg_file.h>
#include "include/bpo_internal.h"
#include <mplib1/daemon.h>
#include <mplib1/bpo.h>
#include <mplib1/fprintfile.h>


/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/*
Queue Test code.
        set stderr.

        :restart_point
        fork()
        parent dies.
        if parent_pid set, wait for parent to die.
        register self.
        set parent_pid = getpid()
        clean process.
        clean queues.
        Check for bad queue list, die if bad.
        Create 'n' queues (some Q_NM_NONE, some Q_NM_PIPE)
        Craete 'n' messages.
        Post messages to queues.
        goto restart_point.

*/


static struct cache_proc_list *cpl=NULL;

/* ------------------------------------------------------------------
    Start of Code
   ------------------------------------------------------------------ */


static int
proc_list( struct shm_process_private *sppp, void *param1 )
{
    int ks;
    ks = is_pid_dead(sppp->pid);
    fprintfile(stderr, "Process %d is %s\n",
		sppp->pid, ((ks)?"dead":"alive") );
    if (ks)
	return(SHM_DELETE_PROC);
    return(SHM_LEAVE_PROC);
}

static void
q_pid( struct bpo_private_pid *ppp, int *alive )
{
    if (is_pid_dead(ppp->pid)==0)
	(*alive)++;
    return;
}

static int
q_list( struct bpo_private_q *pqp, void *param1 )
{
    int rv = Q_LEAVE_QUEUE;

    if (pqp->Q_notify == Q_NM_PIPE)
    {
	if (pqp->Q_flags & Q_SOLE_READER)
	{
	    char tstr[PID_STR_LEN];
	    gen_pid_str( tstr, pqp->Q_sole_pid );
	    if (dl_Find_Item_By_Name( &cpl->proc_list, tstr )==NULL)
		rv = Q_DELETE_QUEUE;
	}else
	{
	    int any_alive;
	    any_alive=0;
	    dl_Walk_List( &pqp->private_pids,
			(dl_Walk_List_t)q_pid, &any_alive );
	    if (any_alive==0)
		rv = Q_DELETE_QUEUE;
	}
    }
    return(rv);
}


static int
comms_cleanup( void *hint )
{
    fprintfile( stderr, "Clean out dead process slots\n" );
    Validate_Processes( hint, SHM_CHECK_ALL_PROCESSES, proc_list, NULL );
    fprintfile( stderr, "Clean out redundant queues\n" );
    cpl = Get_Cache_Procs( hint );
    Validate_Queues( hint, Q_CHECK_ALL_QUEUES, q_list, NULL );
    return(1);
}

static int
do_list_checks( void *hint )
{
    struct Q_Head *qh;
    struct shm_proc_list *splp;

    qh = get_Q_resource( hint );
    bpo_Validate_List( &qh->Qh_List );

    splp = Find_Proc_List( hint );
    bpo_Validate_List(  &splp->proc_List );
    return(0);
}

static int
create_new_queues( void *hint, int p_cnt )
{
    int t;
    char q_name[50];
    struct bpo_Q *qp;
    struct bpo_Q_Mesg *mp;

    for (t=0;t<p_cnt;t++)
    {
	sprintf( q_name, "q_test.%d.%d", (int)getpid(), t );
	qp = mpCreateQ( hint, q_name, 0, Q_SOLE_READER, Q_NM_PIPE );
	mp = shalloc( hint, sizeof(struct bpo_Q_Mesg) );
	if (qp && mp)
	{
	    bpo_Init_Node( &mp->Q_Node, NULL, mp );
	    mpPutMsg( q_name, &mp->Q_Node );
	}else
	{
	    fprintfile(stderr,"Cannot allocate: %p %p\n", qp, mp );
	    if (mp)
		shfree( mp );
	}
    }
    return(0);
}

static int
main_loop( void *hint, int p_cnt )
{
    pid_t parent_pid=(pid_t)0;
    int t;
    int loop_cnt=0;

    while(loop_cnt++<p_cnt)
    {
	fprintfile(stderr,"Loop: %d\n", loop_cnt );
	t = fork();
	if (t>0)
	{
	    /* parent, time to die */
	    fprintfile( stderr, "Parent shutting down\n" );
	    exit(0);
	}else if (t<0)
	{
	    /* problem, time to die */
	    fprintfile( stderr, "Unable to fork, dying\n" );
	    exit(1);
	}
	/* We are child */
	if (parent_pid != (pid_t)0)
	{
	    while( is_pid_dead( parent_pid ) == 0);
	}
	parent_pid = getpid();
	Register_Process_Details( hint );
	comms_cleanup(hint );
	do_list_checks( hint );
	create_new_queues( hint, p_cnt );
    };
    fprintfile(stderr,"Finishing\n" );
    return(0);
}

int
main( int argc, char *argv[] )
{
    void *hint;
    char *shm_name,*e_nm=NULL;
    char *cp=NULL;
    int optc,lpcnt=3;

/*    set_config_string( "STDERR", "q_test.err" );*/
    while ((optc = getopt(argc, argv, "e:c:f:")) != EOF)
    switch(optc)
    {
    case 'e' : /* Specify config file */
	e_nm =optarg;
	break;
    case 'f' : /* Specify config file */
	read_config_file(cp=optarg);
	break;
    case 'c' : /* Count */
	lpcnt=atoi(optarg);
	break;
    default:
/*	usage( argv[0], "Invalid argument supplied" );*/
	break;
    }
    if (cp==NULL)
    {
	cp = eval_config_default( NULL, "$HOME/config/q_test.cfg" );
	read_config_file( cp );
    }
    if (e_nm)
    {
	set_config_string( "STDERR", e_nm );
    }

    shm_name = strdup(eval_config_default( "SHM_CONFIG", "$HOME/config/shm.cfg" ));
    hint = comms_shm_attach( shm_name, SODB_READ_CONFIG );
    fprintfile(stderr,"Attach returned: %p\n", hint );
    if (hint)
    {
	leavehome();
	main_loop(hint, lpcnt);
    }
    return(0);
}

/* -- End of File -- */
